/*
 * This file is part of vospace-file-service
 * Copyright (C) 2021 Istituto Nazionale di Astrofisica
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
package it.inaf.ia2.transfer.service;

import it.inaf.ia2.transfer.auth.TokenPrincipal;
import it.inaf.ia2.transfer.persistence.FileDAO;
import it.inaf.ia2.transfer.persistence.model.FileInfo;
import java.io.File;
import java.io.FileInputStream;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import org.junit.jupiter.api.AfterAll;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.kamranzafar.jtar.TarEntry;
import org.kamranzafar.jtar.TarInputStream;
import static org.mockito.ArgumentMatchers.any;
import org.mockito.Mock;
import static org.mockito.Mockito.when;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.test.util.ReflectionTestUtils;
import org.springframework.util.FileSystemUtils;

@ExtendWith(MockitoExtension.class)
public class ArchiveServiceTest {

    @Mock
    private FileDAO fileDAO;

    @Mock
    private AuthorizationService authorizationService;

    private ArchiveService archiveService;

    private static File tmpDir;

    @BeforeAll
    public static void setUpClass() throws Exception {
        tmpDir = Files.createTempDirectory("generated").toFile();
    }

    @AfterAll
    public static void tearDownClass() throws Exception {
        FileSystemUtils.deleteRecursively(tmpDir);
    }

    @BeforeEach
    public void setUp() {
        archiveService = new ArchiveService(tmpDir.getAbsolutePath());
        ReflectionTestUtils.setField(archiveService, "fileDAO", fileDAO);
        ReflectionTestUtils.setField(archiveService, "authorizationService", authorizationService);
    }

    @Test
    public void testTarGeneration() throws Exception {

        String parent = "/path/to";

        File tmpParent = tmpDir.toPath().resolve("test1").toFile();
        File file1 = createFile(tmpParent, "dir1/a/b/file1");
        File file2 = createFile(tmpParent, "dir1/a/b/file2");
        File file3 = createFile(tmpParent, "dir2/c/file3");
        File file4 = createFile(tmpParent, "dir2/c/file4");
        File file5 = createFile(tmpParent, "dir2/c/d/file5");
        File file6 = createFile(tmpParent, "file6");

        ArchiveJob job = new ArchiveJob();
        job.setPrincipal(new TokenPrincipal("user123", "token123"));
        job.setJobId("abcdef");
        job.setType(ArchiveJob.Type.TAR);
        job.setVosPaths(Arrays.asList(parent + "/dir1", parent + "/dir2", parent + "/file6"));

        when(authorizationService.isDownloadable(any(), any())).thenReturn(true);

        List<FileInfo> fileInfos = new ArrayList<>();
        addFileInfo(fileInfos, parent + "/file6", file6);
        addDirInfo(fileInfos, parent + "/dir1");
        addDirInfo(fileInfos, parent + "/dir1/a");
        addDirInfo(fileInfos, parent + "/dir1/a/b");
        addFileInfo(fileInfos, parent + "/dir1/a/b/file1", file1);
        addFileInfo(fileInfos, parent + "/dir1/a/b/file2", file2);
        addDirInfo(fileInfos, parent + "/dir2");
        addDirInfo(fileInfos, parent + "/dir2/c");
        addFileInfo(fileInfos, parent + "/dir2/c/file3", file3);
        addFileInfo(fileInfos, parent + "/dir2/c/file4", file4);
        addDirInfo(fileInfos, parent + "/dir2/c/d");
        addFileInfo(fileInfos, parent + "/dir2/c/d/file5", file5);

        when(fileDAO.getArchiveFileInfos(any())).thenReturn(fileInfos);

        archiveService.createArchive(job);

        File result = tmpDir.toPath().resolve("user123").resolve("abcdef.tar").toFile();

        assertTrue(result.exists());

        // verify result structure
        List<String> expectedSequence = Arrays.asList("file6", "dir1/", "dir1/a/", "dir1/a/b/",
                "dir1/a/b/file1", "dir1/a/b/file2", "dir2/", "dir2/c/", "dir2/c/file3", "dir2/c/file4",
                "dir2/c/d/", "dir2/c/d/file5");

        int i = 0;
        try ( TarInputStream tis = new TarInputStream(new FileInputStream(result))) {
            TarEntry entry;
            while ((entry = tis.getNextEntry()) != null) {
                assertFalse(i >= expectedSequence.size(), "Found more entries than in expected sequence");
                assertEquals(expectedSequence.get(i), entry.getName());
                if (!entry.isDirectory()) {
                    assertEquals("some data", new String(tis.readAllBytes()));
                }
                i++;
            }
        }

        assertFalse(i < expectedSequence.size(), "Found less entries than in expected sequence");
    }

    private void addFileInfo(List<FileInfo> fileInfos, String vosPath, File file) {
        FileInfo fileInfo = new FileInfo();
        fileInfo.setOsPath(file.getAbsolutePath());
        fileInfo.setVirtualPath(vosPath);
        fileInfos.add(fileInfo);
    }

    private void addDirInfo(List<FileInfo> fileInfos, String vosPath) {
        FileInfo fileInfo = new FileInfo();
        fileInfo.setVirtualPath(vosPath);
        fileInfo.setDirectory(true);
        fileInfos.add(fileInfo);
    }

    private File createFile(File parent, String path) throws Exception {
        parent.mkdir();
        String[] files = path.split("/");
        for (int i = 0; i < files.length; i++) {
            File file = parent.toPath().resolve(files[i]).toFile();
            if (i == files.length - 1) {
                // test os_path different from vos_path
                file.renameTo(file.getParentFile().toPath().resolve(file.getName() + "-renamed").toFile());
                file.createNewFile();
                Files.write(file.toPath(), "some data".getBytes());
                return file;
            } else {
                file.mkdir();
                parent = file;
            }
        }
        throw new IllegalStateException("Files have to be created");
    }
}
